<?php

namespace Drupal\Tests\xnttstrjp\Unit;

use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\xnttstrjp\Plugin\ExternalEntities\FieldMapper\StringJsonPath;

/**
 * The class to test StringJsonPath.
 *
 * @group xnttstrjp
 */
class StringJsonPathTest extends UnitTestCase {

  /**
   * Data provider for testExtractValueFromRawDataUsingMapping().
   */
  public function provideTestExtractValueFromRawDataUsingMapping() {
    return [
      // Direct mapping.
      [32, ['bla' => 4, 'bli' => 32], 'bli'],
      // Constant.
      ['bli', ['bla' => 4, 'bli' => 32], '+bli'],
      // Empty value.
      ['', ['bla' => 4], ''],
      // Static string value.
      ['toto', ['bli' => 4], '"toto"'],
      // Static integer value.
      [42, ['bli' => 4], '+42'],
      [-42, [], '-42'],
      // Concate.
      ["blibla", [], "'bli' . \"bla\""],
      // Basic JSON Path.
      ['2', ['bla' => '2'], '$.bla'],
      ['3', ['bla' => '3'], '$[\'bla\']'],
      // Basic String function.
      ['bliblubla', [], "substr('blobliblubla', 3)"],
      ['bliblu', [], "substr('blobliblubla', 3, 6)"],
      // JSON Path and functions combination.
      [
        'blablabla', [
          'foo' => [
            'bar' => [42, 'bloblibluble'],
          ],
          'baz' => 'bla',
        ],
        "substr(str_replace('obli', 'abla', $.foo.bar.1), 0, strpos($.foo.bar.1, 'blu')) . $.baz",
      ],
    ];
  }

  /**
   * Test value extraction from raw data using given mappings.
   *
   * @dataProvider provideTestExtractValueFromRawDataUsingMapping
   */
  public function testExtractValueFromRawDataUsingMapping($expected, $data, $mapping) {
    $configuration = [];
    $plugin_id = 'x';
    $plugin_definition = [];

    $logger = $this->getMockBuilder(LoggerChannelInterface::class)
      ->disableOriginalConstructor()
      ->getMock();
    $logger->expects($this->any())
      ->method('error')
      ->willReturn(NULL);
    $logger_factory = $this->getMockBuilder(LoggerChannelFactoryInterface::class)
      ->disableOriginalConstructor()
      ->getMock();
    $logger_factory
      ->expects($this->any())
      ->method('get')
      ->with('xnttstrjp')
      ->willReturn($logger);
    $container = new ContainerBuilder();
    $container->set('logger.factory', $logger_factory);
    \Drupal::setContainer($container);

    $entity_type_manager = $this->getMockBuilder(EntityTypeManager::class)
      ->disableOriginalConstructor()
      ->getMock();
    $entity_field_manager = $this->getMockBuilder(EntityFieldManagerInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    $typed_data_manager = $this->getMockBuilder(TypedDataManagerInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    $string_translation = $this->getMockBuilder(TranslationInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    $strjpath = new StringJsonPath($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $entity_field_manager, $typed_data_manager, $string_translation);
    $value = $strjpath->extractValueFromRawDataUsingMapping($data, $mapping);
    $this->assertEquals([$expected], $value, 'Failed with ');
  }

  /**
   * Data provider for testInvalidExtractValueFromRawDataUsingMapping().
   */
  public function provideTestInvalidExtractValueFromRawDataUsingMapping() {
    return [
      // Invalid mapping syntax.
      ['', [], '--'],
      ['', [], 'toto'],
      ['', [], '"toto""'],
      ['', [], '..'],
      // Invalid JSON Path.
      ['', [], '$...'],
      ['', ['bla' => 806], '$.bli'],
      // Invalid function.
      ['', ['bla' => 806], 'bli(42)'],
      // Function with avalid arguments.
      ['', ['bla' => 806], 'substr()'],
    ];
  }

  /**
   * Test value extraction failures.
   *
   * @dataProvider provideTestInvalidExtractValueFromRawDataUsingMapping
   */
  public function testInvalidExtractValueFromRawDataUsingMapping($expected, $data, $mapping) {
    $configuration = [];
    $plugin_id = 'x';
    $plugin_definition = [];
    $entity_type_manager = $this->getMockBuilder(EntityTypeManager::class)
      ->disableOriginalConstructor()
      ->getMock();
    $entity_field_manager = $this->getMockBuilder(EntityFieldManagerInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    $typed_data_manager = $this->getMockBuilder(TypedDataManagerInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    $string_translation = $this->getMockBuilder(TranslationInterface::class)
      // ->disableOriginalConstructor()
      ->getMock();
    // $strjpath =
    new StringJsonPath($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $entity_field_manager, $typed_data_manager, $string_translation);
    // $value = $strjpath->extractValueFromRawDataUsingMapping($data, $mapping);
    $this->markTestIncomplete(
      'This test has not been implemented yet.'
    );
  }

}
